# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pyrtshm/protocol/protocol_v1.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pyrtshm/protocol/protocol_v1.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"pyrtshm/protocol/protocol_v1.proto\"\x9b\x01\n\x05State\x12\x15\n\rproto_version\x18\x01 \x01(\r\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12+\n\x0eoperation_type\x18\x03 \x01(\x0e\x32\x0e.OperationTypeH\x00\x88\x01\x01\x12\x12\n\nseq_number\x18\x04 \x01(\x04\x12\x11\n\x04\x64\x61ta\x18\x05 \x01(\x0cH\x01\x88\x01\x01\x42\x11\n\x0f_operation_typeB\x07\n\x05_data*$\n\rOperationType\x12\x07\n\x03SET\x10\x00\x12\n\n\x06\x44\x45LETE\x10\x01\x62\x06proto3'
)

_OPERATIONTYPE = _descriptor.EnumDescriptor(
  name='OperationType',
  full_name='OperationType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=196,
  serialized_end=232,
)
_sym_db.RegisterEnumDescriptor(_OPERATIONTYPE)

OperationType = enum_type_wrapper.EnumTypeWrapper(_OPERATIONTYPE)
SET = 0
DELETE = 1



_STATE = _descriptor.Descriptor(
  name='State',
  full_name='State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='proto_version', full_name='State.proto_version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='State.key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='operation_type', full_name='State.operation_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='seq_number', full_name='State.seq_number', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='State.data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_operation_type', full_name='State._operation_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_data', full_name='State._data',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=39,
  serialized_end=194,
)

_STATE.fields_by_name['operation_type'].enum_type = _OPERATIONTYPE
_STATE.oneofs_by_name['_operation_type'].fields.append(
  _STATE.fields_by_name['operation_type'])
_STATE.fields_by_name['operation_type'].containing_oneof = _STATE.oneofs_by_name['_operation_type']
_STATE.oneofs_by_name['_data'].fields.append(
  _STATE.fields_by_name['data'])
_STATE.fields_by_name['data'].containing_oneof = _STATE.oneofs_by_name['_data']
DESCRIPTOR.message_types_by_name['State'] = _STATE
DESCRIPTOR.enum_types_by_name['OperationType'] = _OPERATIONTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

State = _reflection.GeneratedProtocolMessageType('State', (_message.Message,), {
  'DESCRIPTOR' : _STATE,
  '__module__' : 'pyrtshm.protocol.protocol_v1_pb2'
  # @@protoc_insertion_point(class_scope:State)
  })
_sym_db.RegisterMessage(State)


# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import pickle
import socket
from queue import Queue
from threading import Thread
from typing import Dict, List, Optional, Tuple

from .metrics import Metrics
from .protocol.protocol_v1_pb2 import OperationType, State


class ForwardThread(Thread):
    packet_queue: Queue
    destinations: List[Tuple[str, int]]
    sock: socket.socket

    def __init__(self, sock, destinations, metrics):
        super().__init__(daemon=True)
        self.destinations = destinations
        self.packet_queue = Queue()
        self.metrics = metrics
        self.socket = sock

    def send(self, data):
        self.packet_queue.put(data)

    def run(self):
        while True:
            msg = self.packet_queue.get()
            if msg is None:
                break
            for destination in self.destinations:
                self.socket.sendto(msg, destination)
                self.metrics.sent_packets += 1

    def stop(self):
        self.packet_queue.put(None)


class SharedMemory(Thread):
    host: str
    port: int
    forward_list: List
    forward_thread: ForwardThread
    socket: Optional[socket.socket]
    states: Dict
    data: Dict
    metrics: Metrics

    def __init__(self, listen: Tuple, forward_nodes: List):
        """
        Build the shared memory.

        :param listen: The tuple with (host, port) to listen.
        :param forward_nodes: The list of tuples to where incoming states
            should be forwarded.
        """
        super(SharedMemory, self).__init__(daemon=True)
        self.host, self.port = listen
        self.socket = None
        self.data = {}
        self.states = {}
        self.metrics = Metrics()

        self.forward_nodes = forward_nodes

        self.should_run = True
        self.sock_bufsize = 1024

    def init_socket(self):
        if self.socket is not None:
            return self.socket
        family = socket.AF_INET6 if ":" in self.host else socket.AF_INET
        self.socket = socket.socket(family, socket.SOCK_DGRAM)
        self.socket.bind((self.host, self.port))
        return self.socket

    def init_forward_thread(self):
        self.forward_thread = ForwardThread(
            self.socket, self.forward_nodes, self.metrics
        )
        self.forward_thread.start()

    def decode_key(self, obj_bytes: bytes) -> object:
        return pickle.loads(obj_bytes)

    def decode_value(self, obj_bytes: bytes) -> object:
        return pickle.loads(obj_bytes)

    def encode_key(self, obj: object) -> bytes:
        return pickle.dumps(obj)

    def encode_value(self, obj: object) -> bytes:
        return pickle.dumps(obj)

    def run(self):
        self.init_forward_thread()
        with self.init_socket():
            while self.should_run:
                try:
                    self.process_msg()
                except socket.error:
                    pass

    def process_msg(self):
        data = self.socket.recv(self.sock_bufsize)
        self.metrics.received_packets += 1
        msg = State()
        msg.ParseFromString(data)
        key = self.decode_key(msg.key)

        current_state = self.states.get(key)
        if current_state and current_state.seq_number > msg.seq_number:
            # Don't override our value with older values.
            self.metrics.out_of_order_count += 1
            return

        self.states[key] = msg
        if msg.operation_type == OperationType.DELETE:
            try:
                del self.data[key]
                self.metrics.forward_key_del += 1
            except KeyError:
                # Delete operation should fail silently if the key doesn't
                # exist. We might have missed the state setting this key,
                # for example.
                self.metrics.delete_unknown_key_count += 1
        else:
            self.data[key] = self.decode_value(msg.data)
            self.metrics.forward_key_set += 1

    def stop(self):
        self.should_run = False
        self.forward_thread.stop()

    def get_next_state(self, key: object, value: object) -> State:
        state = self.states.get(key)
        data = self.encode_value(value)
        if state is not None:
            state.seq_number += 1
            state.data = data
        else:
            state = State(key=self.encode_key(key), seq_number=1, data=data)
            self.states[key] = state
        return state

    def get(self, key, default_value=None):
        return self.data.get(key, default_value)

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key: object, value: object):
        state = self.get_next_state(key, value)
        self.data[key] = value
        self.forward_thread.send(state.SerializeToString())

    def __delitem__(self, key):
        state = self.get_next_state(key, None)
        state.operation_type = OperationType.DELETE
        del self.data[key]
        self.forward_thread.send(state.SerializeToString())



import time
from unittest import TestCase

from pyrtshm import SharedMemory


class ValueObject:
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def __eq__(self, other):
        return other.a == self.a and other.b == self.b and other.c == self.c


class TestBasicForwarding(TestCase):
    def start_nodes(self, n=3, initial_port=5551):
        host = "127.0.0.1"
        all_forwards = [(host, initial_port + i) for i in range(n)]
        nodes = []
        for i in range(n):
            forwards = [
                (host, port)
                for host, port in all_forwards
                if port != initial_port + i
            ]
            nodes.append(SharedMemory((host, initial_port + i), forwards))
        for n in nodes:
            # Force socket initialization and timeout.
            n.init_socket()
            n.socket.settimeout(0.1)
            n.start()
            self.addCleanup(n.stop)
            self.addCleanup(n.join)
        return nodes

    def stop_nodes(self, nodes):
        for n in nodes:
            n.stop()
            n.join()
            n.socket.close()

    def wait_nodes(self, nodes):
        time.sleep(0.1)
        for n in nodes:
            n.stop()

    def values_got_set_and_forwarded(self, nodes, key, value):
        some_node = nodes[0]
        some_node[key] = value

        self.wait_nodes(nodes)
        for n in nodes:
            retries = 3
            while retries > 0:
                if n[key] == value:
                    break
                else:
                    if retries == 0:
                        return False
                    else:
                        retries -= 1
                        time.sleep(0.05)
        return True

    def test_primitive_object_set(self):
        keys = [1, b"bin", "some str", 1.34, (1, 2)]
        values = keys + [{1: "a"}, [3, 2, 1], ValueObject("1", b"x", 123)]
        for key in keys:
            if key == 1:
                continue
            for value in values:
                nodes = self.start_nodes()
                self.assertTrue(
                    self.values_got_set_and_forwarded(nodes, key, value),
                    f"Error setting {key}: {value}",
                )
                self.stop_nodes(nodes)
            return

    def test_set_value(self):
        main, *nodes = self.start_nodes(n=3)
        main[1] = 123

        self.wait_nodes([main] + nodes)
        self.assertEqual(2, main.metrics.sent_packets)
        self.assertEqual(0, main.metrics.forward_key_set)
        for n in nodes:
            self.assertEqual(123, n[1])
            self.assertEqual(1, n.metrics.received_packets)
            self.assertEqual(1, n.metrics.forward_key_set)

    def test_get_default_value_key_not_set(self):
        main, *nodes = self.start_nodes(n=3)
        main[1] = 123

        self.wait_nodes([main] + nodes)
        for n in nodes:
            self.assertEqual(123, n.get(1))
            self.assertEqual(-1, n.get("no-existing-key", -1))

    def test_set_value_multiple_times(self):
        main, *nodes = self.start_nodes(n=3)
        main["x"] = 123
        main["x"] = 333
        main["x"] = 987

        self.wait_nodes([main] + nodes)
        self.assertEqual(2 * 3, main.metrics.sent_packets)
        self.assertEqual(0, main.metrics.forward_key_set)
        for n in nodes:
            self.assertEqual(3, n.states["x"].seq_number)
            self.assertEqual(987, n["x"])
            self.assertEqual(3, n.metrics.received_packets)
            self.assertEqual(3, n.metrics.forward_key_set)

    def test_delete_value(self):
        main, *nodes = self.start_nodes()

        main[1] = 123
        del main[1]

        self.wait_nodes([main] + nodes)

        # Should have send 4 packets: 2 to set, and 2 to del.
        self.assertEqual(2 * 2, main.metrics.sent_packets)
        self.assertEqual(0, main.metrics.forward_key_set)
        self.assertEqual(0, main.metrics.forward_key_del)
        for n in nodes:
            self.assertNotIn("x", n.states)
            self.assertRaises(KeyError, lambda: n[1])
            self.assertEqual(0, n.metrics.sent_packets)
            self.assertEqual(1, n.metrics.forward_key_set)
            self.assertEqual(1, n.metrics.forward_key_del)



class Metrics:
    received_packets: int = 0
    sent_packets: int = 0
    forward_key_set: int = 0
    forward_key_del: int = 0
    lost_packet_count: int = 0
    out_of_order_count: int = 0
    delete_unknown_key_count: int = 0
